local addonName, addon = ...
addonName = addonName:match("%a+"):lower()

local pcall, select, strjoin, tostring, type = pcall, select, strjoin, tostring, type
local format, gsub, match = string.format, string.gsub, string.match
local ipairs_iter, pairs_iter, _G = ipairs(_G), pairs(_G)

local ERROR_TYPE = "bad argument #%d to %q (%s expected, got %s)"

local IsObjectType = getmetatable(UIParent).__index.IsObjectType

local chatFrame, HEADER = DEFAULT_CHAT_FRAME
local AddMessage = chatFrame.AddMessage

local LOR = LibStub('LibOutputResource-1.0')
local FetchGlobalName = LOR.FetchGlobalName
local GetInfo = LOR.GetInfo
local GetSpecial = LOR.GetSpecial
local GetTypeFromChunk = LOR.GetTypeFromChunk

local COLOR_BLUE = GetSpecial('BLUE')
local COLOR_GREEN = GetSpecial('GREEN')
local COLOR_WHITE = GetSpecial('WHITE')

local ERROR = GetSpecial('ERROR')
local NIL = GetSpecial()

local OUTPUT = strjoin("", COLOR_BLUE, "%d", COLOR_GREEN, " ", "%s", " ", COLOR_WHITE, "%s", "|r")
local TABLE_INFO = strjoin("", COLOR_BLUE, "%d", " ", "[", COLOR_GREEN, "%s", " ", COLOR_WHITE, "%s", COLOR_BLUE, "]", " ", "=", COLOR_GREEN, " ", "%s", " ", COLOR_WHITE, "%s", "|r")

local NAME = "%s %s%s"

--[[----------------------------------------------------------------------------
Helpers
------------------------------------------------------------------------------]]
local function GetTableSize(table)
	local size = 0
	for key in pairs_iter, table, nil do
		size = size + 1
	end
	return size, table
end

--[[----------------------------------------------------------------------------
Output
------------------------------------------------------------------------------]]
local function AddBuffer()
	local numMessages = chatFrame:GetNumMessages()
	if numMessages >= 1 then
		local lastMessage = chatFrame:GetMessageInfo(numMessages)
		if lastMessage ~= HEADER then
			HEADER = nil
			if not (match(lastMessage, "\n%s*$") or match(lastMessage, "^%s*$")) then
				AddMessage(chatFrame, " ")
			end
		end
	end
end

local function Output(outputFunc, size, ...)
	local start, maxLines
	if not size then
		size = select('#', ...)
	end

	maxLines = chatFrame:GetMaxLines()
	if size < maxLines - 1 then
		start = 1
		AddBuffer()
	else
		start = size - maxLines + 2
	end

	if size > 0 then
		outputFunc(start, size, ...)
	else
		AddMessage(chatFrame, NIL)
	end
	AddMessage(chatFrame, " ")
end

local function OutputArray(start, size, data)
	for index = start, size do
		AddMessage(chatFrame, format(OUTPUT, index, GetInfo(data[index])))
	end
end

local function OutputSpecial(start, size, message, color)
	if message == nil or message == "NIL" then
		message = NIL
	elseif message == "ERROR" then
		message = ERROR
	elseif color then
		color = color:upper()
		if RAID_CLASS_COLORS[color] then
			color = format("|c%s", (CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS)[color].colorStr)
		else
			local specialColor = GetSpecial(color)
			if specialColor then
				color = specialColor
			else
				color = match(color, "|C%x%x(%x%x%x%x%x%x)") or match(color, "%x%x(%x%x%x%x%x%x)") or match(color, "(%x%x%x%x%x%x)")
				if color then
					color = "|cff" .. color
				end
			end
		end
		if color then
			message = strjoin("", color, message, "|r")
		end
	end
	AddMessage(chatFrame, message)
end

local function OutputTable(start, size, data)
	local index, startKey = 1
	for key in pairs_iter, data, nil do
		if index == start then break end
		index, startKey = index + 1, key
	end
	for key, value in pairs_iter, data, startKey do
		local keyType, keyValue = GetInfo(key)
		AddMessage(chatFrame, format(TABLE_INFO, index, keyType, keyValue, GetInfo(value)))
		index = index + 1
	end
end

local function OutputVarArg(start, size, ...)
	for index = start, size do
		AddMessage(chatFrame, format(OUTPUT, index, GetInfo((select(index, ...)))))
	end
end

--[[----------------------------------------------------------------------------
API
------------------------------------------------------------------------------]]
local validMethod = {
	ipairs		= true,
	metatable	= true,
	pairs		= true,
	parse		= true,
	sortedPairs = true,
	special		= true
}

local __index = {
	chatFrame = chatFrame,

	header = function(...)
		local text = strjoin(" ", tostringall(...))
		if text == "" then
			text = " "
		end
		AddBuffer()
		AddMessage(chatFrame, text)
		HEADER = text
	end,

	ipairs = function(data)
		if type(data) ~= 'table' then
			error(format(ERROR_TYPE, 1, addonName .. ".ipairs", 'table', type(data)), 2)
		end
		Output(OutputArray, #data, data)
	end,

	metatable = function(data)
		local metatable = getmetatable(data)
		if type(metatable) == 'table' then
			Output(OutputTable, GetTableSize(type(metatable.__index) == 'table' and metatable.__index or metatable))
		else
			Output(OutputVarArg, 1, metatable)
		end
	end,

	pairs = function(data)
		if type(data) ~= 'table' then
			error(format(ERROR_TYPE, 1, addonName .. ".pairs", 'table', type(data)), 2)
		end
		Output(OutputTable, GetTableSize(data))
	end,

	parse = function(parameters)
		if type(parameters) ~= 'string' then
			error(format(ERROR_TYPE, 1, addonName .. ".parse", 'string', type(parameters)), 2)
		end
		Output(OutputVarArg, nil, SecureCmdOptionParse(parameters))
	end,

	print = function(...)
		local text = strjoin(" ", tostringall(...))
		if text == "" then
			text = " "
		end
		AddMessage(chatFrame, text)
		HEADER = nil
	end,

	special = function(message, color)
		if message ~= nil and type(message) ~= 'string' then
			error(format(ERROR_TYPE, 1, addonName .. ".special", 'string', type(message)), 2)
		elseif color ~= nil and type(color) ~= 'string' then
			error(format(ERROR_TYPE, 2, addonName .. ".special", 'string', type(color)), 2)
		end
		Output(OutputSpecial, 1, message, color)
	end
}

_G[addonName] = setmetatable(addon, {
	__call = function(self, ...)
		Output(OutputVarArg, nil, ...)
	end,

	__index = __index,
	__metatable = addonName,

	__newindex = function(self, index, value)
		if index == "chatFrame" then
			local originalValue = value
			if value == nil or value == 0 then
				value = DEFAULT_CHAT_FRAME
			else
				local type = type(value)
				if type == 'number' then
					value = _G['ChatFrame' .. value]
				elseif type == 'string' then
					value = _G[value]
				end
			end
			if type(value) ~= 'table' then
				error(format(ERROR_TYPE, 1, addonName .. ".chatFrame", 'table', type(originalValue)), 3)
			elseif not pcall(IsObjectType, value, 'ScrollingMessageFrame') then
				error(addonName .. ".chatFrame must be derived from ScrollingMessageFrame", 3)
			elseif type(value.AddMessage) ~= 'function' then
				error(addonName .. ".chatFrame must have an AddMessage method", 3)
			end
			if chatFrame ~= value then
				chatFrame, __index.chatFrame, HEADER = value, value, nil
				AddMessage = chatFrame.AddMessage
			end
		elseif not __index[index] then
			rawset(self, index, value)
		end
	end
})

--[[----------------------------------------------------------------------------
Slash command support
------------------------------------------------------------------------------]]
local ProcessLuaChunk
do
	local function RunProcess(process, ok, ...)
		if ok then
			return pcall(process, ...)
		end
		return ok, (...)
	end

	ProcessLuaChunk = function(chunk, process)
		local lua, err = loadstring("return " .. chunk, chunk)
		if lua then
			lua, err = RunProcess(process, pcall(lua))
		end
		if not lua then
			Output(OutputSpecial, 1, "ERROR")
			geterrorhandler()(err)
		end
	end
end

--[[----------------------------------------------------------------------------
Slash commands
------------------------------------------------------------------------------]]
local actionInfoCommands = { "IsHarmful%s", "IsHelpful%s", "Get%sInfo" }

_G["SLASH_" .. addonName .. "ActionInfo1"] = "/actioninfo"
SlashCmdList[addonName .. "ActionInfo"] = function(command)
	local action = tonumber(command)
	if not action then
		local frame = GetMouseFocus()
		if type(frame) ~= 'table' or type(frame.GetAttribute) ~= 'function' or frame:GetAttribute('type') ~= 'action' or type(frame.action) ~= 'number' then
			return
		end
		action = frame.action
	end
	
	command = format("GetActionInfo(%s)", action)
	addon.header(command, "=")
	ProcessLuaChunk(command, addon)

	local actionType, actionId = GetActionInfo(action)
	if type(actionType) ~= 'string' or type(actionId) ~= 'number' then return end
	actionType = actionType:gsub("%a", string.upper, 1)

	for index = 1, #actionInfoCommands do
		local command = format(actionInfoCommands[index], actionType)
		if type(_G[command]) == 'function' then
			command = format("%s(%s)", command, actionId)
			addon.header(command, "=")
			ProcessLuaChunk(command, addon)
		end
	end
end

_G["SLASH_" .. addonName .. "IPairs1"] = "/ipairs"
SlashCmdList[addonName .. "IPairs"] = function(command)
	addon.header(format("ipairs(%s)", command ~= "" and command or 'nil'), "=")
	ProcessLuaChunk(command, addon.ipairs)
end

_G["SLASH_" .. addonName .. "Metatable1"] = "/metatable"
SlashCmdList[addonName .. "Metatable"] = function(command)
	addon.header(format("getmetatable(%s)", command ~= "" and command or 'nil'), "=")
	ProcessLuaChunk(command, addon.metatable)
end

_G["SLASH_" .. addonName .. "Output1"] = "/output"
_G["SLASH_" .. addonName .. "Output2"] = "/out"
SlashCmdList[addonName .. "Output"] = function(command)
	addon.header(command ~= "" and command or NIL, "=")
	ProcessLuaChunk(command, addon)
end

_G["SLASH_" .. addonName .. "Pairs1"] = "/pairs"
SlashCmdList[addonName .. "Pairs"] = function(command)
	addon.header(format("pairs(%s)", command ~= "" and command or 'nil'), "=")
	ProcessLuaChunk(command, addon.pairs)
end

_G["SLASH_" .. addonName .. "Parse1"] = "/parse"
SlashCmdList[addonName .. "Parse"] = function(command)
	command = format("%q", GetTypeFromChunk(command, 'string'))
	addon.header(format("SecureCmdOptionParse(%s)", command), "=")
	ProcessLuaChunk(command, addon.parse)
end

_G["SLASH_" .. addonName .. "Print1"] = "/print"
SlashCmdList[addonName .. "Print"] = function(command)
	ProcessLuaChunk(command, addon.print)
end
