local LIB, REVISION = "LibOutputResource-1.0", 0
if not LibStub then error(LIB .. " requires LibStub", 0) end

local lib, oldRevision = LibStub:NewLibrary(LIB, REVISION)
if not lib then return end

local loadstring, pcall, tostring, type, wipe = loadstring, pcall, tostring, type, wipe
local format, gsub, match = string.format, string.gsub, string.match
local string_join, string_lower, string_upper = string.join, string.lower, string.upper
local pairs_iter, _G = pairs(_G)

local special = {
	BLUE	= "|cff88ccff",		-- same as /dump
	ERROR	= "",
	GREEN	= "|cff88ff88",		-- same as /dump
	NIL		= "",
	ORANGE	= "|cffffaa66",
	RED		= "|cffff6666",
	WHITE	= "|cffffffff",
	YELLOW	= "|cffffff66"
}

local COLOR_ORANGE, COLOR_RED, COLOR_YELLOW = special.ORANGE, special.RED, special.YELLOW

local NIL			= COLOR_RED .. "<nil>|r"
local WHITESPACE	= COLOR_RED .. "<whitespace:%d>|r"

local NAME = "%s %s%s"

special.ERROR		= COLOR_RED .. "<error>|r"
special.NIL			= NIL

--[[----------------------------------------------------------------------------
Version bridge
------------------------------------------------------------------------------]]
if oldRevision then
	lib.__void()
end

function lib.__void()
	wipe(lib)
end

--[[----------------------------------------------------------------------------
CaseAgnosticRegEx
------------------------------------------------------------------------------]]
do
	local dualcase = setmetatable({ }, {
		__index = function(self, index)
			local lower, upper = string_lower(index), string_upper(index)
			local result = string_join("", "[", upper, lower, "]")
			self[lower], self[upper] = result, result
			return result
		end
	})

	local stack = { }
	local function Preserve(value)
		local index = #stack + 1
		stack[index] = gsub(value, "%%", "%%%%")
		return "$" .. index
	end

	lib.CaseAgnosticRegEx = setmetatable({ }, {
		_call = function(self, key)
			return self[key]
		end,

		__index = function(self, key)
			local regex = gsub(key, "$d+", Preserve)
			regex = gsub(regex, "%%.d*", Preserve)
			regex = gsub(regex, "%b[]d*", Preserve)
			regex = gsub(regex, "[A-Za-z]", dualcase)
			for index = #stack, 1, -1 do
				regex = gsub(regex, "$" .. index, stack[index], 1)
			end
			wipe(stack)
			self[key] = regex
			return regex
		end
	})
end

--[[----------------------------------------------------------------------------
FetchGlobalName
------------------------------------------------------------------------------]]
local FetchGlobalName
do
	local cache = { }

	function FetchGlobalName(object)
		local name = cache[object]
		if name and _G[name] ~= object then
			cache[object], name = nil, nil
		end
		if not name then
			for key, value in pairs_iter, _G, nil do
				if object == value and type(key) == 'string' then
					cache[object], name = key, key
					break
				end
			end
		end
		return name
	end
end
lib.FetchGlobalName = FetchGlobalName

--[[----------------------------------------------------------------------------
GetRawInfo
------------------------------------------------------------------------------]]
local function GetRawInfo(data)
	local dataType, value, subType, name, color = type(data)
	if dataType == 'string' then
		if match(data, "[%S\n]") then
			value = gsub(gsub(format("%q", data), "\n", "n"), "\124", "\124\124")
		else
			value = format(WHITESPACE, #data)
		end
	elseif dataType == 'table' then
		local ok
		ok, name = pcall(data.GetName, data)
		if ok and name then
			color = _G[name] == data and COLOR_YELLOW or COLOR_RED
		else
			name = FetchGlobalName(data)
			color = name and COLOR_ORANGE
		end
		ok, subType = pcall(data.GetObjectType, data)
		if not ok then
			subType = nil
		end
		value = tostring(data)
		value = match(value, "^%a+: 0*(%x+)$") or value
	elseif dataType == 'number' or dataType == 'boolean' then
		value = tostring(data)
	elseif dataType == 'nil' then
		value = NIL
	else
		name = FetchGlobalName(data)
		if name then
			color = COLOR_ORANGE
		end
		value = tostring(data)
		value = match(value, "^%a+: 0*(%x+)$") or value
	end
	return dataType, value, subType, name, color
end
lib.GetRawInfo = GetRawInfo

--[[----------------------------------------------------------------------------
GetInfo
------------------------------------------------------------------------------]]
lib.GetInfo = function(data)
	local dataType, value, subType, name, color = GetRawInfo(data)
	if subType then
		dataType = subType
	end
	if name then
		value = format("%s %s%s|r", value, color, name)
	end
	return dataType, value
end

--[[----------------------------------------------------------------------------
GetSpecial
------------------------------------------------------------------------------]]
lib.GetSpecial = function(name)
	if type(name) == 'string' then
		name = string_upper(name:trim())
	end
	return special[name] or NIL
end

--[[----------------------------------------------------------------------------
GetTypeFromChunk
------------------------------------------------------------------------------]]
lib.GetTypeFromChunk = function(chunk, desiredType)
	if type(chunk) == 'string' and (desiredType ~= 'string' or not _G[chunk]) then
		local lua = loadstring("return " .. chunk, chunk)
		if lua then
			local ok, result = pcall(lua)
			if ok then
				if type(result) == desiredType then
					return result
				elseif type(result) == 'string' and type(_G[result]) == desiredType then
					return _G[result]
				end
			end
		end
	end
	if type(chunk) == desiredType then
		return chunk
	end
end

--[[----------------------------------------------------------------------------
SortSearchResults
------------------------------------------------------------------------------]]
do
	local lowercase = setmetatable({ }, {
		__mode = 'kv',

		__index = function(self, index)
			local result
			if type(index) == 'string' then
				result = string_lower(index)
			else
				result = string_lower(tostring(index))
			end
			self[index] = result
			return result
		end
	})

	lib.SortSearchResults = function(a, b)
		return lowercase[a] < lowercase[b]
	end
end
